import TermsViewMode = require("sas/dm/businessdata/components/termsview/TermsViewMode");
import BusinessDataPMControllerBase = require("sas/dm/businessdata/controller/BusinessDataPMControllerBase");
import TermsComponentController = require("sas/dm/businessdata/controller/TermsComponentController");
import ObjectManagerTermWrapper = require("sas/dm/businessdata/model/ObjectManagerTermWrapper");
import Snapshot = require("sas/dm/businessdata/model/Snapshot");
import Term = require("sas/dm/businessdata/model/Term");
import TermLockStatusResource = require("sas/dm/businessdata/model/TermLockStatusResource");
import TermRevision = require("sas/dm/businessdata/model/TermRevision");
import TermSummary = require("sas/dm/businessdata/model/TermSummary");
import WorkflowTask = require("sas/dm/businessdata/model/WorkflowTask");
import WorkflowType = require("sas/dm/businessdata/model/WorkflowType");
import BusinessDataUtils = require("sas/dm/businessdata/util/BusinessDataUtils");
import BusyKeys = require("sas/dm/businessdata/util/BusyKeys");
import ErrorDictionary = require("sas/dm/businessdata/util/ErrorDictionary");
import GrantedCapabilitiesCache = require("sas/dm/businessdata/util/GrantedCapabilitiesCache");
import SASLineageUtils = require("sas/dm/businessdata/util/sasLineage/SASLineageUtils");
import Context = require("sas/dm/commons/context/Context");
import DynamicBindable = require("sas/dm/commons/model/base/DynamicBindable");
import ErrorResponse = require("sas/dm/commons/model/ErrorResponse");
import Action = require("sas/dm/commons/util/Action");
import BindableArray = require("sas/dm/commons/util/binding/BindableArray");
import BindingUtils = require("sas/dm/commons/util/binding/BindingUtils");
import Unbindable = require("sas/dm/commons/util/binding/Unbindable");
import MenuAction = require("sas/dm/commons/util/MenuAction");
import CommonSharedProperties = require("sas/dm/commons/util/SharedProperties");
import ValidationResponse = require("sas/dm/commons/model/ValidationResponse");
import StringUtils = require("sas/dm/commons/util/StringUtils");
import ArrayUtils = require("sas/dm/commons/util/ArrayUtils");
import TypeAttribute = require("sas/dm/businessdata/model/TypeAttribute");
import MessageDialog = sas.hc.m.MessageDialog;
import CollectionUtil = require("sas/dm/commons/util/CollectionUtil");
import OpenTerm = require("sas/dm/businessdata/model/OpenTerm");
import CustomAttribute = require("sas/dm/businessdata/model/CustomAttribute");

class EditTermPMController extends BusinessDataPMControllerBase {

    public static VIEWEDITBUTTON_KEY_VIEW:string = "view";
    public static VIEWEDITBUTTON_KEY_EDIT:string = "edit";
    public static ALREADY_LOCKED_EVENT:string = "locked";

    // private properties
    private _openTermsTermWatcher: Unbindable;
    private _openTermsRawTermWatcher: Unbindable;
    private _masterControllerWatcher: Unbindable;
    private _dirtyWatcher: Unbindable;
    private _workflowTaskWatcher: Unbindable;
    private _rawTermEditModeWatcher: Unbindable;

    // public properties
    private _term:Term;
    private _rawTerm:Term;
    private _hasEditTermsCapabilities:boolean;
    private _hasEditTermContentCapabilities:boolean;
    private _hasUnlockOtherUserTermsCapability:boolean;
    private _isEditButtonEnabled:boolean;
    private _isEditButtonVisible:boolean;
    private _currentUser:string;
    private _hasNotifyContactsCapability:boolean;
    private _termsViewMode: TermsViewMode;
    private _selectedSnapshot: Snapshot;
    private _sasLineageUtils: SASLineageUtils;
    private _isRevision:boolean;

    private _publishAction:Action;
    private _submitWorkflowAction: Action;

    private _duplicateTermAction:MenuAction;
    private _workflowActionsAction:MenuAction; //this has a submenu...actionception
    private _unlockTermAction:MenuAction;
    private _renameTermAction:MenuAction;
    private _notifyContactsTermAction:MenuAction;
    private _savePdfTermAction:MenuAction;
    private _refreshTermAction:MenuAction;

    private readonly _editTermActionsMenu: Array<MenuAction>;

    public get term(): Term {
        return this._term;
    }

    public set term(value: Term) {
        if (!this.__currentlyHandlingOpenTermsTermChange && value) {
            this.getMasterController().updateOpenTerm(value.id, value, null, true);
        }
        this._term = value;
        BindingUtils.unbind(this._dirtyWatcher);
        this._dirtyWatcher = BindingUtils.handlePropertyChange(this.term, "isDirty", $.proxy(this.handleIsDirtyChanged, this));
    }

    private handleIsDirtyChanged(newValue: boolean, oldValue: boolean): void {
        if (newValue !== oldValue) {
            this.updateActions();
        }
    }

    private handleWorkflowTaskChanged(newValue: WorkflowTask, oldValue: WorkflowTask): void {
        if (newValue != oldValue) {
            if (this.term) {
                if (this.rawTerm) {
                    this.term.workflowTask = this.rawTerm.workflowTask;
                } else {
                    this.term.workflowTask = null;
                }
            }
            this.updateActions();
        }
    }

    public get rawTerm(): Term {
        return this._rawTerm;
    }

    private __settingRawTerm: boolean = false;
    public set rawTerm(value: Term) {
        if (this._rawTerm == value) {
            return;
        }
        this.__settingRawTerm = true;
        try {
            this._rawTerm = value;
            if (!this.__currentlyHandlingOpenTermsRawTermChange && value) {
                this.getMasterController().updateOpenTerm(value.id, null, value, true);
            }

            BindingUtils.unbind(this._workflowTaskWatcher);
            this._workflowTaskWatcher = BindingUtils.handlePropertyChange(this.rawTerm, "workflowTask", $.proxy(this.handleWorkflowTaskChanged, this), true);
            if (this.rawTerm && this.rawTerm.workflowInstanceID && !this.rawTerm.workflowTask && !this.rawTerm.isSpecificRevision) {
                this.getCurrentWorkflowTasks(this.rawTerm, false);
            }

            BindingUtils.unbind(this._rawTermEditModeWatcher);
            this._rawTermEditModeWatcher = BindingUtils.handlePropertyChange(this.rawTerm, "isEditMode", $.proxy(this.onRawTermEditModeChanged, this), true);

        } finally {
            this.__settingRawTerm = false;
        }
    }

    public get hasEditTermsCapabilities(): boolean {
        return this._hasEditTermsCapabilities;
    }

    public set hasEditTermsCapabilities(value: boolean) {
        this._hasEditTermsCapabilities = value;
    }

    public get hasEditTermContentCapabilities(): boolean {
        return this._hasEditTermContentCapabilities;
    }

    public set hasEditTermContentCapabilities(value: boolean) {
        this._hasEditTermContentCapabilities = value;
    }

    public get hasUnlockOtherUserTermsCapability(): boolean {
        return this._hasUnlockOtherUserTermsCapability;
    }

    public set hasUnlockOtherUserTermsCapability(value: boolean) {
        this._hasUnlockOtherUserTermsCapability = value;
    }

    public get isEditButtonEnabled(): boolean {
        return this._isEditButtonEnabled;
    }

    public set isEditButtonEnabled(value: boolean) {
        this._isEditButtonEnabled = value;
    }

    public get isEditButtonVisible(): boolean {
        return this._isEditButtonVisible;
    }

    public set isEditButtonVisible(value: boolean) {
        this._isEditButtonVisible = value;
    }

    public get currentUser(): string {
        return this._currentUser;
    }

    public set currentUser(value: string) {
        this._currentUser = value;
    }

    public get hasNotifyContactsCapability(): boolean {
        return this._hasNotifyContactsCapability;
    }

    public set hasNotifyContactsCapability(value: boolean) {
        this._hasNotifyContactsCapability = value;
    }

    public get termsViewMode(): TermsViewMode {
        return this._termsViewMode;
    }

    public set termsViewMode(value: TermsViewMode) {
        this._termsViewMode = value;
    }

    public get selectedSnapshot(): Snapshot {
        return this._selectedSnapshot;
    }

    public set selectedSnapshot(value: Snapshot) {
        this._selectedSnapshot = value;
    }

    /**
     * If looking to edit, need to attempt to lock term.  If locks successfully, go into edit mode.  If fails, check if
     * it is already locked.  If not, do not let edit as term was not successfully locked.  If already locked, check who
     * it is locked by.  If current user, go into edit mode.  If not current user, do nothing, as it cannot be edited.
     * If looking to view, check if the term is locked.  If locked and not by the current user, do not unlock but do go
     * into view mode.  All other cases unlock and go ino view mode.
     * @param {boolean} newValue
     * @param {boolean} oldValue
     */
    public onRawTermEditModeChanged(newValue: boolean, oldValue: boolean) {
        //even though this may get changed back, apply the change first so that the SegmentedButton control actually
        // updates...if the editMode value doesn't change, the control will show edit despite the value not changing
        // (DJL)
        if(!this.rawTerm){
            return;
        }
        if ((newValue && !oldValue) || (newValue && this.rawTerm.isEditMode && this.rawTerm.lockedBy == null)) {
            this.rawTerm.updateIsBusy(true, BusyKeys.LOCK_TERM);
            //if trying to edit, lock and verify not already locked
            this.getMasterController().lockTerm(this.rawTerm, (result: TermLockStatusResource) => {
                // We don't want to make a midtier call just to update these properties, so manually update them from
                // the TermLockStatusResource on both term and rawTerm.
                this.syncTermLockStatus(result);

                this.rawTerm.updateIsBusy(false, BusyKeys.LOCK_TERM);
            }, (error: ErrorResponse) => {
                if (error.errorCode == ErrorDictionary.TERM_ALREADY_LOCKED) {
                    this.rawTerm.updateIsBusy(true, BusyKeys.GET_TERM_LOCKED_STATUS);
                    this.rawTerm.updateIsBusy(false, BusyKeys.LOCK_TERM);
                    this.getMasterController().getTermLockedStatus(this.rawTerm, (result: TermLockStatusResource) => {
                        this.rawTerm.updateIsBusy(false, BusyKeys.GET_TERM_LOCKED_STATUS);
                        if (result.lockedById != sas.hc.ui.core.util.UserInfo.instance.userid) {
                            this.rawTerm.isEditMode = false;
                            super.fireEvent(EditTermPMController.ALREADY_LOCKED_EVENT, {lockedByName: result.lockedByName});
                        }
                    }, (error: ErrorResponse) => {
                        this.rawTerm.isEditMode = false;
                        this.handleError(error,BusyKeys.GET_TERM_LOCKED_STATUS);
                        this.rawTerm.updateIsBusy(false,BusyKeys.GET_TERM_LOCKED_STATUS);
                    });
                } else {
                    this.rawTerm.isEditMode = false;
                    this.handleError(error,BusyKeys.LOCK_TERM);
                    this.rawTerm.updateIsBusy(false,BusyKeys.LOCK_TERM);
                }
            });
        }
    }

    public get viewEditButtonKeyView():string {
        return EditTermPMController.VIEWEDITBUTTON_KEY_VIEW;
    }

    public get viewEditButtonKeyEdit():string {
        return EditTermPMController.VIEWEDITBUTTON_KEY_EDIT;
    }

    public get publishAction(): Action {
        return this._publishAction;
    }

    public set publishAction(value: Action) {
        this._publishAction = value;
    }

    public get submitWorkflowAction(): Action {
        return this._submitWorkflowAction;
    }

    public set submitWorkflowAction(value: Action) {
        this._submitWorkflowAction = value;
    }

    public get isRevision(): boolean {
        return this._isRevision;
    }

    public set isRevision(value: boolean) {
        this._isRevision = value;
    }

    public get duplicateTermAction(): MenuAction {
        return this._duplicateTermAction;
    }

    public set duplicateTermAction(value: MenuAction) {
        this._duplicateTermAction = value;
    }

    public get workflowActionsAction(): MenuAction {
        return this._workflowActionsAction;
    }

    public set workflowActionsAction(value: MenuAction) {
        this._workflowActionsAction = value;
    }

    public get unlockTermAction(): MenuAction {
        return this._unlockTermAction;
    }

    public set unlockTermAction(value: MenuAction) {
        this._unlockTermAction = value;
    }

    public get renameTermAction(): MenuAction {
        return this._renameTermAction;
    }

    public set renameTermAction(value: MenuAction) {
        this._renameTermAction = value;
    }

    public get notifyContactsTermAction(): MenuAction {
        return this._notifyContactsTermAction;
    }

    public set notifyContactsTermAction(value: MenuAction) {
        this._notifyContactsTermAction = value;
    }

    public get savePdfTermAction(): MenuAction {
        return this._savePdfTermAction;
    }

    public set savePdfTermAction(value: MenuAction) {
        this._savePdfTermAction = value;
    }

    public get refreshTermAction(): MenuAction {
        return this._refreshTermAction;
    }

    public set refreshTermAction(value: MenuAction) {
        this._refreshTermAction = value;
    }

    public get editTermActionsMenu(): Array<MenuAction> {
        return this._editTermActionsMenu;
    }

    public set editTermActionsMenu(value:Array<MenuAction>){
        (<BindableArray>this._editTermActionsMenu).replace(value);
    }

    constructor() {
        super();
        this._hasUnlockOtherUserTermsCapability = false;
        this._isEditButtonEnabled = false;
        this._isEditButtonVisible = false;
        this._isRevision = false;
        this._hasEditTermsCapabilities = false;
        this._hasEditTermContentCapabilities = false;
        this._hasNotifyContactsCapability = false;

        this.duplicateTermAction = new MenuAction(
            this.resources.getText("terms.createEditPage.bar.button.saveas.title"), () => {}, null,
            this.resources.getText("terms.createEditPage.bar.button.saveas.title"),
            false, true);
        this.workflowActionsAction = new MenuAction(
            this.resources.getText("terms.viewactions.workflowactions.menuitem.label"), null, null,
            this.resources.getText("terms.viewactions.workflowactions.tip.label"),
            true, true);
        this.unlockTermAction = new MenuAction(
            this.resources.getText("terms.viewactions.menuitem.unlock.title"), () => {}, null,
            this.resources.getText("terms.viewactions.menuitem.unlock.title"),
            true, false, false);
        this.renameTermAction = new MenuAction(
            this.resources.getText("terms.viewactions.menuitem.rename.title"), () => {}, null,
            this.resources.getText("terms.viewactions.menuitem.rename.title"),
            false, false);
        this.notifyContactsTermAction = new MenuAction(
            this.resources.getText("terms.viewactions.menuitem.notifyall.title"), () => {}, null,
            this.resources.getText("terms.viewactions.menuitem.notifyall.title"),
            true, true);
        this.savePdfTermAction = new MenuAction(
            this.resources.getText("terms.viewactions.menuitem.saveaspdf.title"), () => {}, null,
            this.resources.getText("terms.viewactions.menuitem.saveaspdf.title"),
            true, false);
        this.refreshTermAction = new MenuAction(
            this.resources.getText("terms.viewactions.menuitem.refresh.title"), () => {}, null,
            this.resources.getText("terms.viewactions.menuitem.refresh.title"),
            true, true);
        this.workflowActionsAction.subActions = new BindableArray();
        this._editTermActionsMenu = new BindableArray([this.duplicateTermAction, this.workflowActionsAction, this.unlockTermAction, this.renameTermAction, this.notifyContactsTermAction, this.savePdfTermAction, this.refreshTermAction]);
    }

    protected getMasterController():TermsComponentController {
        return <TermsComponentController>super.getMasterController();
    }

    protected postSetContext(value:Context):void {
        if(this.getMasterController() != null){
            // set up watchers, etc...
            this.initialize();
        }
        this.getView().setModel(this.asBindable(), "editTerm");
    }

    private initialize():void {
        this.unbind();

        this.hasEditTermsCapabilities = GrantedCapabilitiesCache.getInstance().hasCapability(GrantedCapabilitiesCache.EDIT_TERMS);
        this.hasEditTermContentCapabilities = GrantedCapabilitiesCache.getInstance().hasCapability(GrantedCapabilitiesCache.EDIT_TERM_CONTENTS);
        this.hasNotifyContactsCapability = GrantedCapabilitiesCache.getInstance().hasCapability(GrantedCapabilitiesCache.NOTIFY_CONTACTS);
        this.hasUnlockOtherUserTermsCapability = GrantedCapabilitiesCache.getInstance().hasCapability(GrantedCapabilitiesCache.UNLOCK_ANOTHER_USERS_TERMS);
        this.currentUser = sas.hc.ui.core.util.UserInfo.instance.userid;

        //isVisible and isEnabled set in updateActions
        this.publishAction = new Action(
            this.resources.getText("terms.createEditPage.bar.button.publish.title"),
            $.proxy(this.saveTermAndPublish, this),
            sas.icons.HC.PUBLISH,
            this.resources.getText("terms.createEditPage.bar.button.publish.title"));

        this.submitWorkflowAction = new Action(
          this.resources.getText("terms.createEditPage.bar.button.submit.title"),
            $.proxy(this.submitWorkflow, this),
            sas.icons.HC.SENDTOWORKFLOW,
            null);

        this._masterControllerWatcher = BindingUtils.handlePropertiesChange(this.getMasterController(), ["termsViewMode", "selectedSnapshot", "activeDrilledInTerm"],
            (property: string, newValue: any, oldValue: any) => {
                if (property == "termsViewMode") {
                    this.termsViewMode = <TermsViewMode>newValue;
                } else if (property == "selectedSnapshot") {
                    this.selectedSnapshot = <Snapshot>newValue;
                } else if (property == "activeDrilledInTerm") {
                    if (TermSummary.EqualsById(this.rawTerm, newValue) || (!this.rawTerm && !!newValue)) {
                        this.rawTerm = <Term>newValue;
                        this.updateActions();
                    }
                }
            }, true);

        this._openTermsTermWatcher = BindingUtils.handleArrayItemPropertyChange(this.getMasterController().openTerms, "term",
            $.proxy(this.onOpenTermsTermChanged, this),
            true);

        this._openTermsRawTermWatcher = BindingUtils.handleArrayItemPropertyChange(this.getMasterController().openTerms, "rawTerm",
            $.proxy(this.onOpenTermsRawTermChanged, this),
            true);

        //using any so getItem() can be called...this seems to be a subjective name set on ObjectManager.openItem()
        //that doesn't have a universal getter.
        if ((<any>this.byId("editTermPage")).getItem() && (<any>this.byId("editTermPage")).getItem() instanceof ObjectManagerTermWrapper) {
            const termWrapper = <ObjectManagerTermWrapper>((<any>this.byId("editTermPage")).getItem());
            // set rawTerm to the original term from the tree so we can still monitor it for changes.
            this.rawTerm = <Term>termWrapper.term;
            if(termWrapper.term instanceof TermRevision){
                this.isRevision = true;
            }
        }
        this.updateActions();
        this._sasLineageUtils = SASLineageUtils.getInstance();
    }

    private __currentlyHandlingOpenTermsTermChange:boolean = false;
    private onOpenTermsTermChanged(item: OpenTerm, property: string, newValue: Term, oldValue: Term): void {
        // Note: this.term might not be set yet, so grab the id from this.rawTerm
        if (newValue && !item.isInSelectedTermPanel && item.representsSameTerm(this.rawTerm) && newValue != this.term) {
            this.__currentlyHandlingOpenTermsTermChange = true;
            try {
                this.term = newValue;
                this.updateActions();
            } finally {
                this.__currentlyHandlingOpenTermsTermChange = false;
            }
        }
    }

    private __currentlyHandlingOpenTermsRawTermChange:boolean = false;
    private onOpenTermsRawTermChanged(item: OpenTerm, property: string, newValue: Term, oldValue: Term): void {
        if (newValue && item.representsSameTerm(this.rawTerm) && newValue != this.rawTerm) {
            this.__currentlyHandlingOpenTermsRawTermChange = true;
            try {
                this.rawTerm = newValue;
                this.updateActions();
            } finally {
                this.__currentlyHandlingOpenTermsRawTermChange = false;
            }
        }
    }

    public cleanupBeforeTermClose(closeFunction: () => void): void {
        let openTerm:OpenTerm = this.getMasterController().getOpenTermForTerm(this.term);
        if (!openTerm) {
            if (closeFunction) {
                closeFunction();
            }
            return;
        }
        if (openTerm.rawTerm && openTerm.rawTerm.isEditMode) {
            this.doUnlockTerm(openTerm.rawTerm,()=>{
                this.getMasterController().removeOpenTermForTerm(openTerm.rawTerm);
            });
        }else {
            //TODO if rawTerm is null then openTerm will not be removed from the openTerms array
            this.getMasterController().removeOpenTermForTerm(openTerm.rawTerm);
        }

        if (closeFunction) {
            closeFunction();
        }
    }

    public doUnlockTerm(term:Term = null, unlockSuccessHandler: () => void = null):void {
        if (!term) {
            if (!this.rawTerm) {
                return;
            } else {
                term = this.rawTerm;
            }
        }
        //check if locked by current user, and if so unlock
        this.getMasterController().getTermLockedStatus(term, (result:TermLockStatusResource)=>{
            if (result.lockedById == this.currentUser) {
                //locked by current user
                this.unlockTerm(term, unlockSuccessHandler);
            } else if (result.lockedById == null) {
                // not locked, just continue on...
                this.syncTermLockStatus(result);

                //S1508492:  even if we didn't actually have to call the midtier to do an unlock, we still need
                //to ensure that the edit mode is set to false since the view depends on being in view mode after
                //finding out that the term is not locked
                term.isEditMode = false;

                this.updateActions();
                if (unlockSuccessHandler) {
                    unlockSuccessHandler();
                }
            }
        }, (error: ErrorResponse) => {
            // do nothing
        });
    }

    private syncTermLockStatus(termLockStatus: TermLockStatusResource): void {
        if (this.term) {
            this.term.lockedBy = termLockStatus.lockedById;
            this.term.lockedByName = termLockStatus.lockedByName;
            this.term.lockedTimeStamp = termLockStatus.lockedDate;
        }
    }

    public unlockTerm(term:Term = null, successHandler: () => void = null): void {
        if (!term) {
            if (!this.rawTerm) {
                return;
            } else {
                term = this.rawTerm;
            }
        }
        term.updateIsBusy(true, BusyKeys.UNLOCK_TERM);
        this.getMasterController().unlockTerm(term, (result: TermLockStatusResource) => {
            if (this.rawTerm && TermSummary.EqualsById(this.rawTerm, term)) {
                this.term = <Term>this.rawTerm.clone(this.term);
            }
            term.isEditMode = false;
            this.updateActions();

            term.updateIsBusy(false, BusyKeys.UNLOCK_TERM);

            if (successHandler) {
                successHandler();
            }
        }, (error: ErrorResponse) => {
            term.handleError(error, BusyKeys.UNLOCK_TERM);
        });
    }

    public refreshTerm(successHandler: Function): void {
        this.term.updateIsBusy(true, BusyKeys.GET_TERM);
        let isEditMode: boolean = this.rawTerm.isEditMode;
        this.getMasterController().getTerm(this.term.id, this.rawTerm,
            (response: Term) => {
                this.term.updateIsBusy(false, BusyKeys.GET_TERM);
                this.rawTerm.isEditMode = isEditMode;
                this.updateActions();
                if (this.rawTerm.workflowInstanceID && !this.rawTerm.isSpecificRevision) {
                    this.getCurrentWorkflowTasks(this.rawTerm, true);
                }
                if (successHandler) {
                    successHandler();
                }
            }, (error: ErrorResponse) => {
                this.term.handleError(error, BusyKeys.GET_TERM);
            });
    }

    public resetTerm(): void {
        if (this.rawTerm.isEditMode) {
            if (this.term.isDirty) {
                this.term.isDirty = false;
            }
            this.doUnlockTerm();
        }
    }

    public getCurrentWorkflowTasks(term: Term | TermSummary, refreshCurrent: boolean = false): void {
        if (term && (!term.workflowTask || refreshCurrent === true) && !this.selectedSnapshot) {
            this.getMasterController().getCurrentWorkflowTasks([term], (response: Array<WorkflowTask>) => {
                // dawong - 27 Mar 2019
                // S1495789 - The first time you get a workflow task for a term
                // after its workflow ends, the most recent version of the term
                // in the database still has a workflowInstanceID.  That
                // workflowInstanceID is set to null in the UI in
                // TermsComponentController.updateWorkflowTasksForTerms(), but
                // the actions still need to be updated here since this is
                // normally triggered when the workflowTask changes, not when
                // it's set from null to null.
                this.updateActions();
            }, (error: ErrorResponse) => {
                term.handleError(error, null);
            });
        }
    }

    private startWorkflow(): void {
        this.term.updateIsBusy(true, BusyKeys.START_WORKFLOW);
        this.getMasterController().startWorkflow(true, [this.rawTerm],
            this.rawTerm.hasPublished ? WorkflowType.EDIT : WorkflowType.CREATE,
            $.proxy(this.onSubmitWorkflowSuccess, this),
            (error: ErrorResponse) => {
                this.term.handleError(error, BusyKeys.START_WORKFLOW);
            }
        );
    }

    public submitWorkflow(): void {
        // Check first if the term is a draft
        if (this.term.isDirty) {
            this.saveTermAsDraft( ()=> {
                this.doUnlockTerm(this.rawTerm, $.proxy(this.startWorkflow, this));
            });
        } else {
            // dawong - 22 Aug 2019
            // S1528901 - A term without changes still needs to be validated
            // before submitting it, since an invalid term can be saved in
            // some scenarios.  For example, if a type has a required attribute
            // with no default value, a new term of that type will be invalid
            // when it's first created.
            const response: ValidationResponse<any> = this.validateTermDraft();
            if (response.isValid) {
                this.doUnlockTerm(this.rawTerm, $.proxy(this.startWorkflow, this));
            }
        }
    }

    private onSubmitWorkflowSuccess(): void {
        // raw term should have been updated, so update edit term to equal rawterm
        this.term = <Term>this.rawTerm.clone(this.term);
        this.updateActions();
        this.term.updateIsBusy(false, BusyKeys.START_WORKFLOW);
    }

    //onSuccess/onError are callbacks for ObjectManager when closing unsaved object
    public saveTermAsDraft(onSuccess?:Function, onError?:Function):void {
        this.term.updateIsBusy(true, BusyKeys.SAVE_TERM_AS_DRAFT);
        const response: ValidationResponse<any> = this.validateTermDraft();
        if (response.isValid) {
            this.term.isDirty = false;
            this.getMasterController().saveTermAsDraft(this.term, (result: Term) => {
                this.term.updateIsBusy(false, BusyKeys.SAVE_TERM_AS_DRAFT);
                this.doUnlockTerm(this.rawTerm, () => {
                    if (onSuccess) {
                        onSuccess();
                    }
                });
            }, (error: ErrorResponse) => {
                if (error.errorCode != CommonSharedProperties.ERROR_CODE_CANCELLED) {
                    if (error.errorCode == ErrorDictionary.NO_ROWS_UPDATED) {
                        this.term.updateIsBusy(false, BusyKeys.SAVE_TERM_AS_DRAFT);
                        this.term.isDirty = false;
                        MessageDialog.information({
                            text: this.resources.getText("info.cant.save.unchanged.draft.fmt.msg", [this.rawTerm.name])
                        });
                    } else {
                        this.term.isDirty = true; // Saved failed -- It's still dirty!
                        this.term.handleError(error, BusyKeys.SAVE_TERM_AS_DRAFT);
                    }

                    if (onError) {
                        onError();
                    }
                }
            });
        } else {
            if (onError) {
                onError();
            }
            this.term.updateIsBusy(false, BusyKeys.SAVE_TERM_AS_DRAFT);
        }
    }

    public saveTermAndPublish():void {
        this.term.updateIsBusy(true, BusyKeys.SAVE_TERM_AND_PUBLISH);
        const response: ValidationResponse<any> = this.validateTermDraft();
        if (response.isValid) {
            if (this.rawTerm.isEditMode && this.term.isDirty) {
                this.term.isDirty = false;
                this.getMasterController().saveTermAndPublish(this.term, $.proxy(this.onSavePublishSuccessHandler, this), $.proxy(this.onSavePublishErrorHandler, this, true));
            } else {
                //in view mode or term isn't dirty, just publish
                this.getMasterController().publishTerms([this.term.id], (responseMap: DynamicBindable) => {
                    let result:Term|ErrorResponse = responseMap[this.term.id];
                    if (result && result instanceof ErrorResponse) {
                        this.onSavePublishErrorHandler(false, result);
                    } else if (result && result instanceof Term) {
                        this.getMasterController().updateReferencesToTerm(result);
                        this.onSavePublishSuccessHandler(result);
                    } else {
                        this.term.updateIsBusy(false, BusyKeys.SAVE_TERM_AND_PUBLISH); //shouldn't hit this
                    }
                }, $.proxy(this.onSavePublishErrorHandler, this, false));
            }
        } else {
            this.term.updateIsBusy(false, BusyKeys.SAVE_TERM_AND_PUBLISH);
        }
    }

    private onSavePublishSuccessHandler(result:Term):void {
        this.term.updateIsBusy(false, BusyKeys.SAVE_TERM_AND_PUBLISH);

        this.updateActions();
        this.rawTerm.isEditMode = false; //switch to view mode
    }

    private onSavePublishErrorHandler(wasDirty: boolean, error: ErrorResponse): void {
        this.term.isDirty = wasDirty;
        if(error.errorCode == CommonSharedProperties.ERROR_CODE_CANCELLED){
            this.term.updateIsBusy(false, BusyKeys.SAVE_TERM_AND_PUBLISH);
            return;
        }
        if (error.errorCode == ErrorDictionary.NOT_LOCKED_BY_ACCOUNT) {
            error.message = this.resources.getText("error.cant.publish.locked.term.fmt.msg", [this.rawTerm.name]);
        } else if (error.errorCode == ErrorDictionary.PARENT_TERM_NOT_PUBLISHED) {
            error.message = this.resources.getText("error.cant.publish.draft.parent.fmt.msg", [this.rawTerm.name]);
        }
        else if (error.errorCode == ErrorDictionary.NO_ROWS_UPDATED) {
            this.term.updateIsBusy(false, BusyKeys.SAVE_TERM_AND_PUBLISH);
            this.term.isDirty = false;
            MessageDialog.information({
                text: this.resources.getText("info.cant.publish.unchanged.term.fmt.msg", [this.rawTerm.name])
            });
            return; //FYI: with return statement here this.term.handleError will not be called, which is what we want
        }
        this.term.handleError(error, BusyKeys.SAVE_TERM_AND_PUBLISH);
    }

    public updateActions(): void {
        this.duplicateTermAction.isVisible = this.hasEditTermsCapabilities;

        this.submitWorkflowAction.isVisible = this.hasEditTermsCapabilities && BusinessDataUtils.showWorkflowSubmit(this.term, this.currentUser);
        this.submitWorkflowAction.isEnabled = BusinessDataUtils.enableWorkflowSubmit(this.term, this.currentUser);

        this.unlockTermAction.isVisible = this.hasEditTermContentCapabilities || this.hasUnlockOtherUserTermsCapability;
        this.unlockTermAction.isEnabled = !!this.rawTerm && !this.rawTerm.isEditMode && !!this.term && !!this.term.lockedBy && ((this.term.lockedBy === this.currentUser) || this.hasUnlockOtherUserTermsCapability);

        this.renameTermAction.isVisible = this.hasEditTermsCapabilities;
        this.renameTermAction.isEnabled = this.hasEditTermsCapabilities && BusinessDataUtils.enableRenameTerm(this.term);

        this.notifyContactsTermAction.isVisible = this.hasNotifyContactsCapability;

        this.savePdfTermAction.isEnabled = !!this.rawTerm && !this.rawTerm.isEditMode;

        this.publishAction.isVisible = !BusinessDataUtils.isWorkflowControlledTerm(this.term) && this.hasEditTermsCapabilities;
        this.publishAction.isEnabled = BusinessDataUtils.enablePublishTerm(this.term, this.currentUser);

        this.isEditButtonEnabled = !!this.rawTerm && this.hasEditTermContentCapabilities && BusinessDataUtils.enableEditTerm(this.rawTerm, this.currentUser);
        this.isEditButtonVisible = this.isEditButtonEnabled || (!!this.rawTerm && !!this.rawTerm.workflowTask && this.rawTerm.workflowTask.claimable && !BusinessDataUtils.isDeleteWorkflowActive(this.rawTerm));

        this.updateWorkflowActions();
    }

    private updateWorkflowActions(): void {
        this.workflowActionsAction.subActions.splice(0, this.workflowActionsAction.subActions.length);

        let claimVisibleForActiveTerm: boolean = false;
        let unclaimVisibleForActiveTerm: boolean = false;
        if (this.rawTerm) {
            // if the term doesn't have a workflow task, claim and unclaim won't be visible
            if (this.rawTerm.workflowTask) {
                this.rawTerm.workflowTask.claimAction.isEnabled = this.rawTerm.workflowTask.claimable;
                this.rawTerm.workflowTask.unclaimAction.isEnabled = this.rawTerm.workflowTask.isActualOwner;
                BusinessDataUtils.calculateWorkflowPanelVisibility(this.rawTerm.workflowTask);
            }

            this.submitWorkflowAction.isVisible = this.hasEditTermContentCapabilities && BusinessDataUtils.showWorkflowSubmit(this.rawTerm, this.currentUser);
            if (this.rawTerm.workflowTask) {
                claimVisibleForActiveTerm = this.rawTerm.workflowTask.claimAction.isVisible;
                unclaimVisibleForActiveTerm = this.rawTerm.workflowTask.unclaimAction.isVisible;
                this.workflowActionsAction.subActions.push(this.rawTerm.workflowTask.claimAction);
                this.workflowActionsAction.subActions.push(this.rawTerm.workflowTask.unclaimAction);
            }
            if (this.submitWorkflowAction.isVisible) {
                this.workflowActionsAction.subActions.push(this.submitWorkflowAction);
            }
        } else {
            this.submitWorkflowAction.isVisible = false;
        }

        // only enable the workflow actions submenu if the current user can
        // claim or unclaim the current tasks or can submit the term to workflow
        this.workflowActionsAction.isEnabled = claimVisibleForActiveTerm || unclaimVisibleForActiveTerm || this.submitWorkflowAction.isVisible;

        if (this.workflowActionsAction.isEnabled && this.rawTerm.workflowTask) {
            let firstAction:boolean = true;
            for (let action of this.rawTerm.workflowTask.workflowTransitionActions) {
                if (firstAction) {
                    action.startsSection = true;
                    firstAction = false;
                }
                this.workflowActionsAction.subActions.push(action);
            }
        }
    }

    public validateTermDraft(): ValidationResponse<any>{
        let validationResponse: ValidationResponse<any> = new ValidationResponse<any>(true, "");

        if (this.term) {
            const attributes = this.term.attributes;
            const typeAttributes = this.term.termType.attributes;
            if (!ArrayUtils.isEmpty(attributes) && !ArrayUtils.isEmpty(typeAttributes)) {
                //find the first invalid attribute value
                attributes.forEach(attr => {
                    let items = CollectionUtil.findItemsByField(typeAttributes, "label", attr.label);
                    if(items && items.length>0) {
                        let typeAttribute: TypeAttribute = <TypeAttribute>items[0];
                        if (validationResponse.isValid && typeAttribute.isRequired && StringUtils.isEmpty(attr.value)) {
                            validationResponse.isValid = false;
                            validationResponse.message = this.resources.getText("error.missing.attribute.value.fmt.msg", [attr.label]);
                        }
                    } else {
                        validationResponse.isValid = false;
                        validationResponse.message = this.resources.getText("error.termtype.not.found.message");
                    }
                });
            }
            // if the standard attributes are good, then check the custom attributes
            if (validationResponse.isValid) {
                const customAttributes: Array<CustomAttribute> = this.term.customAttributes;
                let length: number = customAttributes.length;
                for (let i: number = 0; i < length; i++) {
                    if (customAttributes[i].isRequired && StringUtils.isEmpty(customAttributes[i].value)) {
                        validationResponse.isValid = false;
                        validationResponse.message = this.resources.getText("error.missing.attribute.value.fmt.msg", [customAttributes[i].label]);
                        break;
                    }
                }
            }
        }

        if (!validationResponse.isValid) {
            MessageDialog.error({
                text: validationResponse.message
            });
        }
        return validationResponse;
    }

    public renameTerm(newTermName: string, successHandler: () => void) {
        this.getMasterController().renameTerm(this.rawTerm, newTermName, () => {
            this.term.name = this.rawTerm.name;
            if (successHandler) {
                successHandler();
            }
        });
    }

    public duplicateTerm(term: Term, versionLabel: string, successHandler: (response: Term) => void): void {
        const response: ValidationResponse<any> = this.validateTermDraft();
        if (response.isValid) {
            const hasUnsavedChanges: boolean = (this.rawTerm.isEditMode && this.term.isDirty);
            const originalTerm = this.rawTerm;
            if (hasUnsavedChanges) {
                let clonedTerm: Term = Term.createFromSummary(this.term, this.term);
                clonedTerm.name = term.name;
                clonedTerm.parentId = term.parentId;

                this.getMasterController().duplicateAndUpdateTerm(clonedTerm, versionLabel, (duplicatedTerm: Term) => {
                    successHandler(duplicatedTerm);
                    // The term was in a dirty state, clear the dirty flag prior to ending edit mode and closing,
                    // otherwise the user will be prompted about unsaved changes.
                    this.term.isDirty = false;
                    this.rawTerm.isEditMode = false;
                    duplicatedTerm.isEditMode = true;
                    this.setDrilledInTerm(duplicatedTerm);
                    this.getMasterController().termToClose = originalTerm;
                }, (error: ErrorResponse) => {
                    this.handleError(error, null);
                });
            } else {
                this.getMasterController().duplicateTerm(term, versionLabel, (duplicatedTerm: Term) => {
                    successHandler(duplicatedTerm);

                    this.rawTerm.isEditMode = false;
                    duplicatedTerm.isEditMode = true;
                    this.setDrilledInTerm(duplicatedTerm);
                    this.getMasterController().termToClose = originalTerm;
                }, (error: ErrorResponse) => {
                    this.handleError(error, null);
                });
            }
        }
    }

    public termRefresh(): void {
        this.refreshTerm(null);
    }

    public setDrilledInTerm(term: Term) {
        this.getMasterController().activeDrilledInTerm = term;
    }

    public getDrilledInTerm(): Term {
        return this.getMasterController().activeDrilledInTerm;
    }

    public unbind():void {
        BindingUtils.unbind(this._openTermsTermWatcher);
        BindingUtils.unbind(this._openTermsRawTermWatcher);
        BindingUtils.unbind(this._masterControllerWatcher);
        BindingUtils.unbind(this._dirtyWatcher);
        BindingUtils.unbind(this._workflowTaskWatcher);
        BindingUtils.unbind(this._rawTermEditModeWatcher);
    }

    public onViewLineage():void {
        if (this.term) {
            this._sasLineageUtils.onViewLineage(this.term);
        }
    }

    public mayEditTermWithErrorDialog():boolean{
        return BusinessDataUtils.mayEditTermWithErrorDialog(this.resources, this.rawTerm, this.currentUser);
    }

    public mayRenameTermWithErrorDialog():boolean{
        return BusinessDataUtils.mayRenameTermWithErrorDialog(this.resources, this.rawTerm, this.currentUser);
    }

    public destroy():void{
        super.destroy();
        this.unbind();
    }
}

export = EditTermPMController;
